package backup.local;

import index.IndexPanel;

import java.awt.Cursor;
import java.awt.Toolkit;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.ProgressMonitor;
import javax.swing.SwingWorker;
import javax.swing.filechooser.FileFilter;

import allrounder.DbHandler;
import allrounder.StringLib;
import allrounder.TaggedImage;
import allrounder.md5.Hasher;

public class Import {
	/**
	 * Searches the Images (Files) for the hashes in the parameter. if bImport is false, non-reconstructable images
	 * are deleted from DB (after user is asked for export)
	 * @param imagesToImport the TaggedImage-Images with hash and, if bImport is false, ID
	 * @param panel The IndexPanel to set the Cursor and update the filepanel
	 * @param forbiddenFolders The folders in which should not be searched for images. null for no exclusions
	 * @param bImport True if it is an Import, False if it is an attempt to reconstruct lost images
	 * @return false if user aborted (could mean that the submitted images have to be deleted from the db manually)
	 */
	public static void searchFilesForHashes(ArrayList<TaggedImage> imagesToImport, IndexPanel panel, ArrayList<File> forbiddenFolders, boolean bImport) {
		// "convert" to final to get them to work in the anonymous inner class
		final IndexPanel pnl = panel;
		final ArrayList<TaggedImage> images = imagesToImport;
		final ArrayList<File> forbFolders = forbiddenFolders;
		final int iImagesToReconstructCount = images.size();
		final ArrayList<TaggedImage> notReconstructionableImages = new ArrayList<TaggedImage>();
		final boolean bIsImport = bImport;
		
		final StringLib lib = StringLib.getInstance();
		
		pnl.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		
		final ProgressMonitor progressMonitor = new ProgressMonitor(pnl,
                lib.getValue("backup_local.Import.progress.text"),
                "", 0, 100);
		progressMonitor.setProgress(0);
		
		final SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {
			@Override
			protected Void doInBackground() throws Exception {
				// get all images which are present in the selected directories
				ArrayList<File> files = new ArrayList<File>();
		    	TreeMap<File, Boolean> folders = DbHandler.getInstance().getDirectories();
				Iterator<File> folderIter = folders.keySet().iterator();
				while (folderIter.hasNext()) {
					File folder = folderIter.next();
					if (forbFolders != null && forbFolders.contains(folder)) continue;
					for (File image : getImagesFromFolder(folder)) {
						files.add(image);
					}
				}
				folders = null;
				
				int iNumberOfImagesInFolders = files.size(); // this are about 90%
				double dProgress = 0;
				TreeMap<String, ArrayList<File>> hashImgMap = new TreeMap<String, ArrayList<File>>();
				for (File image : files) {
					if (!image.exists()) continue;
					if (isCancelled()) {
						return null;
					}
					String hash = Hasher.imageToHash(image);
					if (hashImgMap.get(hash) != null) {
						hashImgMap.get(hash).add(image);
					} else {
						ArrayList<File> tmp = new ArrayList<File>();
						tmp.add(image);
						hashImgMap.put(hash, tmp);
					}
					dProgress += 90.0/iNumberOfImagesInFolders;
					if ((int)dProgress > getProgress()) setProgress((int)dProgress);
				}
				files = null;
				System.gc();
				setProgress(90);
				dProgress = 90;
				
				if (isCancelled()) {
					return null;
				}
				for (TaggedImage img : images) {
					ArrayList<File> imgPaths = hashImgMap.get(img.hash);
					if (imgPaths == null) {
						notReconstructionableImages.add(img);
						continue;
					}
					
					if (bIsImport) {
						for (File f : imgPaths) {
							img.path = f;
							img.changed = false; // always for imports
							DbHandler.getInstance().addImageToDatabase(img);
							pnl.getFilePanel().setFileStateIndexedNoUpdate(f);
						}
					} else {
						for (File f : imgPaths) {
							DbHandler.getInstance().changePathOfImgInDb(img.ID, f);
							pnl.getFilePanel().setFileStateIndexedNoUpdate(f);
						}
					}
					dProgress += 10.0/iImagesToReconstructCount;
					if ((int)dProgress == 100) {
						setProgress(99);
					} else if ((int)dProgress > getProgress()) {
						setProgress((int)dProgress);
					}
				}
				pnl.getFilePanel().updateList();
				
				// ask for export, then delete
				if (!bIsImport && notReconstructionableImages.size() != 0) {
					int result2 = JOptionPane.showConfirmDialog(pnl, lib.getValue("backup_local.Import.nonrecdel.text", notReconstructionableImages.size() + " ", ""), lib.getValue("backup_local.Import.nonrecdel.title"), JOptionPane.YES_NO_OPTION);
					if (result2 == JOptionPane.YES_OPTION) {
						Export.doExport(notReconstructionableImages, pnl);
					}
					for (TaggedImage img : notReconstructionableImages) {
						if (img.path != null) pnl.getFilePanel().setFileStateNotIndexedNoUpdate(img.path);
						DbHandler.getInstance().removeImageFromDatabase(img.ID);
					}
				}
				setProgress(100);
				hashImgMap = null;
				System.gc();
				return null;
			}
		};
		
		worker.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent evt) {
				if ("progress" == evt.getPropertyName() ) {
					int progress = (Integer) evt.getNewValue();
					progressMonitor.setProgress(progress);
		            String message = String.format(lib.getValue("backup_local.Import.progress.note"), progress);
		            progressMonitor.setNote(message);
		            if (progressMonitor.isCanceled() || worker.isDone()) {
		                Toolkit.getDefaultToolkit().beep();
		                if (progressMonitor.isCanceled()) {
		                    worker.cancel(true);
		                } else {
		                	JOptionPane.showMessageDialog(pnl, (iImagesToReconstructCount - notReconstructionableImages.size()) + "/" + iImagesToReconstructCount + " " + lib.getValue("backup_local.Import.recreated.text"), lib.getValue("backup_local.Import.recreated.title"), JOptionPane.INFORMATION_MESSAGE);
		        			System.gc();
		                }
		                pnl.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		            }
				}
			}
		});
		worker.execute();
		
		return;// (!worker.isCancelled());
	}
	
	public static void doImport(IndexPanel indexPanel) {
		final StringLib lib = StringLib.getInstance();
		
		JFileChooser chooser = new JFileChooser();
		chooser.setFileFilter(new FileFilter() {
			@Override
			public boolean accept(File arg0) {
				if (arg0.getName().endsWith("." + lib.getValue("backup_local.filefilter.extension")) ||
						arg0.isDirectory()) {
					return true;
				}
				return false;
			}
			@Override
			public String getDescription() {
				return lib.getValue("backup_local.filefilter.descr");
			}
		});
		int returnVal = chooser.showOpenDialog(null);
		
	    if(returnVal != JFileChooser.APPROVE_OPTION) return;
	    File imFileName = chooser.getSelectedFile();
	    if(!imFileName.exists() || imFileName.isDirectory()) return;
	    
	    int allImport = JOptionPane.showConfirmDialog(null, lib.getValue("backup_local.Import.allImport.text"), lib.getValue("backup_local.Import.allImport.title"), JOptionPane.YES_NO_CANCEL_OPTION);
		if (allImport != JOptionPane.YES_OPTION && allImport != JOptionPane.NO_OPTION) return;
		boolean bImportAll = (allImport == JOptionPane.YES_OPTION);
	    
		ArrayList<TaggedImage> images = new ArrayList<TaggedImage>();
	    try {
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(imFileName));
			double importversion = ois.readDouble();
			double realversion = Double.parseDouble(DbHandler.getInstance().getGeneral("Version"));
			int sizeOfImgs = ois.readInt(); // save the size for importing
			for (int i = 0; i < sizeOfImgs; ++i) {
				TaggedImage img = (TaggedImage)ois.readObject();
				if (realversion != importversion) {
					// the import file was written with an older version of YiffIndex
					if (importversion == 1.0) {
						// former spezies 112 (unicorn) moved to species 97 (species 97-111 +1)
						TreeSet<Integer> newSpeciesSet = new TreeSet<Integer>();
						for (int spez : img.spezies) {
							if (spez < 97) {
								newSpeciesSet.add(spez);
								continue;
							}
							if (spez == 112) {
								newSpeciesSet.add(97);
							} else {
								newSpeciesSet.add(spez + 1);
							}
						}
						
						// new species 113 (elephant), species 114 (rhino) and species -24 (odd-toed ungulates)
						// nothing to do, species added at the end
						img.spezies = newSpeciesSet;
						importversion = 1.01;
					} // note for future klaue: do not add an else so YI stays backwards compatible
					if (importversion >= 1.01 && importversion <= 1.12) {
						// nothing to do
						importversion = 1.13;
					} // note for future klaue: do not add an else so YI stays backwards compatible
				}
				
				images.add(img);
			}
			ois.close();
		} catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, lib.getValue("backup_local.Export.serializeerror.text") + e.getLocalizedMessage(), StringLib.getInstance().getValue("backup_local.Export.serializeerror.title"), JOptionPane.ERROR_MESSAGE);
		}
		
		if (!bImportAll) {
			//System.out.println("NOT ALL IMPORT");
			// find out which images are allready indexed
			ArrayList<TaggedImage> filterArr = new ArrayList<TaggedImage>();
			TreeSet<TaggedImage> set = DbHandler.getInstance().getAllImages();
			for (TaggedImage img : images) {
				boolean bImgIsInDb = false;
				for (TaggedImage imgInDb : set) {
					if (imgInDb.hash.equals(img.hash)) {
						bImgIsInDb = true;
						break;
					}
				}
				if (!bImgIsInDb) filterArr.add(img);
			}
			images = filterArr;
		}
		
		searchFilesForHashes(images, indexPanel, null, true);
	}
	
	static ArrayList<File> getImagesFromFolder (File folder) {
		ArrayList<File> imagePaths = new ArrayList<File>();
		if (!folder.isDirectory()) return imagePaths;
		File[] files = folder.listFiles();
		if (files.length == 0) return imagePaths;
		for (File f : files) {
			String path = f.getAbsolutePath().toLowerCase();
			// only GIF, JPEG, TIFF, PNG, PICT, Photoshop, BMP, Targa, ICO, CUR,
			// Sunraster, XBM, XPM, and PCX. see com.sun.jimi.core.decoder.*
			if (path.endsWith(".apf") || path.endsWith(".bmp") ||
					path.endsWith(".cur") || path.endsWith(".gif") ||
					path.endsWith(".ico") || path.endsWith(".pcx") ||
					path.endsWith(".pict") || path.endsWith(".png") ||
					path.endsWith(".psd") || path.endsWith(".sunraster") ||
					path.endsWith(".tga") || path.endsWith(".tiff") ||
					path.endsWith(".xbm") || path.endsWith(".xpm") ||
					path.endsWith(".jpg") || path.endsWith(".jpeg") ||
					path.endsWith(".tif")) {
				imagePaths.add(f);
			}
		}
		return imagePaths;
	}
}